# Copyright (c) Alibaba, Inc. and its affiliates.
from .pose_datasource import PoseDataSourceForVideoRec
from .video_datasource import VideoDatasource
from .video_text_datasource import VideoTextDatasource
