# Copyright (c) Alibaba, Inc. and its affiliates.
from .det_mobilenet_v3 import OCRDetMobileNetV3
from .det_resnet_vd import OCRDetResNet
from .rec_mobilenet_v3 import OCRRecMobileNetV3
from .rec_mv1_enhance import OCRRecMobileNetV1Enhance
from .rec_svtrnet import SVTRNet
