# xuebinqin/U-2-Net is licensed under the

# Apache License 2.0
# A permissive license whose main conditions require preservation of copyright and license notices. Contributors provide an express grant of patent rights. Licensed works, modifications, and larger works may be distributed under different terms and without source code.

# Permissions
#  Commercial use
#  Modification
#  Distribution
#  Patent use
#  Private use