# Copyright (c) Alibaba, Inc. and its affiliates.
from . import data_sources, pipelines  # pylint: disable=unused-import
from .ocr_cls_dataset import OCRClsDataset
from .ocr_det_dataset import OCRDetDataset
from .ocr_rec_dataset import OCRRecDataset
